# -*- coding: utf-8 -*-
"""
  Copyright 2022 Mitchell Isaac Parker

   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.

"""

from ..functions.lst import res_to_lst, type_lst, str_to_lst
from ..functions.table import mask_equal, mask_search, order_rows, mask_matrix
from ..functions.col import (
    pdb_code_col,
    chainid_col,
    bb_resid_col,
    edia_col,
    complete_col,
)
from ..functions.path import save_table, save_matrix


def add_edia_status(
    df,
    edia_dict,
    edia_min=0.4,
    edia_resids=None,
    edia_atomids=None,
):
    if edia_atomids is None:
        edia_atomids = "O"

    edia_resid_lst = res_to_lst(edia_resids)
    edia_atomid_lst = type_lst(edia_atomids)

    for index in list(df.index.values):

        pdb_code = df.at[index, pdb_code_col]
        chainid = df.at[index, chainid_col]

        if edia_resid_lst is None:
            index_resid_lst = str_to_lst(df.at[index, bb_resid_col])
        else:
            index_resid_lst = list()
            for resid in edia_resid_lst:
                index_resid_lst.append(resid)

        edia_status = "Not Available"
        edia_below = False
        if pdb_code in list(edia_dict.keys()):
            if chainid in list(edia_dict[pdb_code].keys()):
                edia_status = f"Above {edia_min}"
                for index_resid in index_resid_lst:
                    if edia_below:
                        break
                    if index_resid in list(edia_dict[pdb_code][chainid].keys()):
                        for atomid in edia_atomid_lst:
                            if atomid in list(
                                edia_dict[pdb_code][chainid][index_resid].keys()
                            ):
                                edia_score = edia_dict[pdb_code][chainid][index_resid][
                                    atomid
                                ][edia_col]
                                if edia_score < edia_min:
                                    edia_below = True
                                    break

        if edia_below:
            edia_status = edia_status.replace("Above", "Below")

        df.at[index, edia_col] = edia_status

    return df


def mask_dih_table(df):

    fit_df = mask_equal(df, complete_col, str(True), reset_index=False)

    if edia_col in list(fit_df.columns):
        fit_df = mask_search(
            fit_df, edia_col, "Above", " ", equal=True, reset_index=False
        )

    pred_df = df.loc[~df.index.isin(list(fit_df.index.values)), :]

    return fit_df, pred_df


def mask_dih_data(
    df,
    matrix,
    fit_table_path,
    fit_matrix_path,
    pred_table_path=None,
    pred_matrix_path=None,
    edia_dict=None,
    edia_min=None,
    edia_resids=None,
    edia_atomids=None,
):

    if edia_dict is not None and pdb_code_col in list(df.columns):
        df = add_edia_status(
            df,
            edia_dict,
            edia_min=edia_min,
            edia_resids=edia_resids,
            edia_atomids=edia_atomids,
        )

    fit_df, pred_df = mask_dih_table(df)

    fit_df = order_rows(fit_df, reset_index=False)
    pred_df = order_rows(pred_df, reset_index=False)

    fit_index_lst = list(fit_df.index.values)
    pred_index_lst = list(pred_df.index.values)

    fit_df = fit_df.reset_index(drop=True)
    save_table(fit_table_path, fit_df)

    fit_matrix = mask_matrix(matrix, fit_index_lst, fit_index_lst)
    save_matrix(fit_matrix_path, fit_matrix)

    if pred_table_path is not None:
        pred_df = pred_df.reset_index(drop=True)
        save_table(pred_table_path, pred_df)

    if pred_matrix_path is not None:
        pred_matrix = mask_matrix(matrix, pred_index_lst, fit_index_lst)
        save_matrix(pred_matrix_path, pred_matrix)

    print("Masked dihedral data!")



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['rascore',
 'rascore.util.PDBrenum',
 'rascore.util.PDBrenum.src.download',
 'rascore.util.PDBrenum.src.renum.PDB',
 'rascore.util.PDBrenum.src.renum.mmCIF',
 'rascore.util.PDBrenum.src.renum.shared',
 'rascore.util.constants',
 'rascore.util.functions',
 'rascore.util.pages',
 'rascore.util.pipelines',
 'rascore.util.scripts']

package_data = \
{'': ['*'],
 'rascore': ['.streamlit/*',
             'util/*',
             'util/data/*',
             'util/data/rascore_cluster/*',
             'util/data/rascore_cluster/SW1/*',
             'util/data/rascore_cluster/SW1/Disordered.0P/*',
             'util/data/rascore_cluster/SW1/Disordered.2P/*',
             'util/data/rascore_cluster/SW1/Disordered.3P/*',
             'util/data/rascore_cluster/SW1/Y32in.0P/*',
             'util/data/rascore_cluster/SW1/Y32in.2P/*',
             'util/data/rascore_cluster/SW1/Y32in.3P/*',
             'util/data/rascore_cluster/SW1/Y32out.0P/*',
             'util/data/rascore_cluster/SW1/Y32out.2P/*',
             'util/data/rascore_cluster/SW1/Y32out.3P/*',
             'util/data/rascore_cluster/SW2/*',
             'util/data/rascore_cluster/SW2/Disordered.0P/*',
             'util/data/rascore_cluster/SW2/Disordered.2P/*',
             'util/data/rascore_cluster/SW2/Disordered.3P/*',
             'util/data/rascore_cluster/SW2/Y71in.0P/*',
             'util/data/rascore_cluster/SW2/Y71in.2P/*',
             'util/data/rascore_cluster/SW2/Y71in.3P/*',
             'util/data/rascore_cluster/SW2/Y71out.0P/*',
             'util/data/rascore_cluster/SW2/Y71out.2P/*',
             'util/data/rascore_cluster/SW2/Y71out.3P/*'],
 'rascore.util.PDBrenum': ['src/*', 'src/renum/*']}

install_requires = \
['Bio==1.3.3',
 'Pillow==9.0.1',
 'lxml==4.6.5',
 'matplotlib==3.3.4',
 'matplotlib_venn==0.11.6',
 'numpy==1.21',
 'pandas==1.3.5',
 'pendulum==2.1.2',
 'py3Dmol==1.8.0',
 'pyfiglet==0.8.post1',
 'rdkit_pypi==2021.9.5.1',
 'requests==2.25.1',
 'scipy==1.6.2',
 'seaborn==0.11.1',
 'statannot==0.2.3',
 'statsmodels==0.13.1',
 'stmol==0.0.7',
 'streamlit==1.8.1',
 'tqdm==4.59.0']

entry_points = \
{'console_scripts': ['rascore = rascore.rascore_cli:cli']}

setup_kwargs = {
    'name': 'rascore',
    'version': '1.0.0',
    'description': 'A tool for analyzing RAS protein structures',
    'long_description': None,
    'author': 'mitch-parker',
    'author_email': 'mip34@drexel.edu',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.8,<3.9',
}


setup(**setup_kwargs)



# -*- coding: utf-8 -*-
"""
  Copyright 2022 Mitchell Isaac Parker

   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.

"""

import pandas as pd
from tqdm import tqdm
import concurrent.futures

from ..functions.seq import calc_seq_id, load_record_lst, get_record_seq
from ..functions.lst import lst_to_str, res_to_lst, lst_nums, str_to_lst
from ..functions.coord import load_coord, get_resname, has_resid
from ..functions.col import (
    core_path_col,
    modelid_col,
    chainid_col,
    seq_col,
    mut_status_col,
    mut_pos_col,
    uniprot_id_col,
)
from ..functions.table import get_df_at_index, fix_val
from ..functions.path import save_table, get_seq_path
from ..functions.url import uniprot_url
from ..functions.download import download_file


def build_mut_df(df, index, uniprot_dict, resid_lst=None, coord_path_col=None):

    if coord_path_col is None:
        coord_path_col = core_path_col

    df = get_df_at_index(df, index)

    coord_path = df.at[index, coord_path_col]
    modelid = df.at[index, modelid_col]
    chainid = df.at[index, chainid_col]

    structure = load_coord(coord_path)

    uniprot_acc_lst = list(uniprot_dict.keys())

    mut_status_dict = dict()
    mut_pos_dict = dict()
    uniprot_id_lst = list()

    for uniprot_acc in uniprot_acc_lst:

        mut_status_dict[uniprot_acc] = list()
        mut_pos_dict[uniprot_acc] = list()

        uniprot_seq = uniprot_dict[uniprot_acc][seq_col]

        coord_seq = ""
        for resid in lst_nums(1, len(uniprot_seq)):
            resname = "-"
            if has_resid(structure, chainid, resid, modelid=modelid):
                if resname != "X":
                    resname = get_resname(
                        structure[fix_val(modelid, return_int=True)][chainid][resid],
                        letter=True,
                    )
                    ref = uniprot_dict[uniprot_acc][resid]

                    if resname != ref:
                        if resid is not None:
                            add_mut = True
                            if resid_lst is not None:
                                if resid not in resid_lst:
                                    add_mut = False
                            if add_mut:
                                mut_pos = f"{ref}{resid}"
                                mut_status_dict[uniprot_acc].append(mut_pos + resname)
                                mut_pos_dict[uniprot_acc].append(mut_pos)

            coord_seq += resname

        uniprot_id_lst.append(calc_seq_id(coord_seq, uniprot_seq, aln=False))

    uniprot_acc = uniprot_acc_lst[uniprot_id_lst.index(max(uniprot_id_lst))]

    mut_status_lst = mut_status_dict[uniprot_acc]
    mut_pos_lst = mut_pos_dict[uniprot_acc]

    df.at[index, mut_status_col] = lst_to_str(mut_status_lst, empty="WT")
    df.at[index, mut_pos_col] = lst_to_str(mut_pos_lst, empty="WT")
    df.at[index, uniprot_id_col] = uniprot_acc

    return df


def annot_mut(
    df,
    uniprot_accs,
    mut_table_path=None,
    resids=None,
    seq_dir=None,
    coord_path_col=None,
    num_cpu=1,
):

    df = df.reset_index(drop=True)

    if type(uniprot_accs) == list:
        uniprot_acc_lst = uniprot_accs
    else:
        uniprot_acc_lst = str_to_lst(uniprot_accs, sep_txt=" ")

    if resids is not None:
        resid_lst = res_to_lst(resids)
    else:
        resid_lst = None

    uniprot_dict = dict()

    for uniprot_acc in uniprot_acc_lst:

        uniprot_dict[uniprot_acc] = dict()

        fasta_url = f"{uniprot_url}{uniprot_acc}.fasta"

        fasta_file = get_seq_path(uniprot_acc, dir_path=seq_dir)

        download_file(fasta_url, fasta_file)

        for record in load_record_lst(fasta_file):

            seq = get_record_seq(record)

        uniprot_dict[uniprot_acc][seq_col] = seq

        for i, resname in enumerate(seq):

            uniprot_dict[uniprot_acc][i + 1] = resname

    mut_df = pd.DataFrame()

    if num_cpu == 1:
        for index in tqdm(
            list(df.index.values), desc="Annotating mutations", position=0, leave=True
        ):
            mut_df = pd.concat(
                [
                    mut_df,
                    build_mut_df(
                        df,
                        index,
                        uniprot_dict,
                        resid_lst=resid_lst,
                        coord_path_col=coord_path_col,
                    ),
                ],
                sort=False,
            )
    else:
        with concurrent.futures.ProcessPoolExecutor(max_workers=num_cpu) as executor:
            job_lst = [
                executor.submit(
                    build_mut_df,
                    df,
                    index,
                    uniprot_dict,
                    resid_lst=resid_lst,
                )
                for index in list(df.index.values)
            ]

            for job in tqdm(
                concurrent.futures.as_completed(job_lst),
                desc="Annotating mutations",
                total=len(job_lst),
                miniters=1,
                position=0,
                leave=True,
            ):

                mut_df = pd.concat([mut_df, job.result()], sort=False)

    mut_df = mut_df.reset_index(drop=True)

    print("Annotated mutations!")

    if mut_table_path is not None:
        save_table(mut_table_path, mut_df)
    else:
        return mut_df



from src.download.modules import *


def delete_outdated_files(files_to_refresh, path_to_the_files):
    os.chdir(path_to_the_files)
    for file_for_removal in files_to_refresh:
        try:
            os.remove(file_for_removal)
        except FileNotFoundError:
            pass
    os.chdir(current_directory)


"""usage of delete_outdated_files"""
# delete_outdated_files(refresher_for_mmCIF, format_to_delete = "mmCIF")
# delete_outdated_files(refresher_for_PDB, format_to_delete = "PDB")
# delete_outdated_files(refresher_for_SIFTS, format_to_delete = "SIFTS")



# -*- coding: utf-8 -*-
"""
  Copyright 2022 Mitchell Isaac Parker

   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.

"""

import pandas as pd
import numpy as np
from tqdm import tqdm
import itertools
from sklearn.cluster import DBSCAN
from sklearn.metrics import silhouette_samples

from ..functions.table import mask_matrix, mask_unequal, build_col_count_dict, lst_col
from ..functions.lst import (
    calc_simpson,
    lst_nums,
    build_range_lst,
    sort_lst,
    lst_unique,
    get_lst_val_indices,
)
from ..functions.col import (
    cluster_col,
    silh_col,
    simi_col,
    total_complete_col,
    cluster_count_col,
    total_cluster_pre_col,
    total_cluster_post_col,
    total_noise_pre_col,
    total_noise_post_col,
    total_pruned_nn_col,
    total_pruned_constr_col,
    select_col,
)
from ..functions.cluster import (
    merge_clusters,
    build_cutoff_dict,
    prune_cluster_members,
    build_sum_table,
)
from ..functions.path import save_table


def run_dbscan(matrix, eps, min_samples):

    return (
        DBSCAN(eps=eps, min_samples=min_samples, metric="precomputed")
        .fit(matrix)
        .labels_
    )


def build_pass_lst(label_lst, matrix, max_dist, min_silh):

    pass_lst = list()

    cluster_lst = list(set(label_lst))

    if len(cluster_lst) > 1:

        silhouette_lst = silhouette_samples(matrix, label_lst, metric="precomputed")

        for cluster in cluster_lst:
            if cluster != -1:

                cluster_index_lst = [
                    index for index, label in enumerate(label_lst) if label == cluster
                ]

                max_cluster_dist = np.max(
                    mask_matrix(matrix, cluster_index_lst, cluster_index_lst)
                )
                mean_cluster_silh = np.mean(silhouette_lst[cluster_index_lst])

                if max_cluster_dist <= max_dist and mean_cluster_silh > min_silh:
                    pass_lst.append(tuple([label_lst, cluster_index_lst]))

    return pass_lst


def check_connect(compare_lst, i_index, j_index, min_simi):

    overlap = calc_simpson(compare_lst[i_index][1], compare_lst[j_index][1])

    if overlap >= min_simi:
        status = 1

    if overlap < min_simi:
        status = 0

    return tuple([i_index, j_index, status])


def connect_graph(graph_dict, node_lst, result_dict):

    for node in node_lst:

        subgraph = set(
            [k for k, v in list(graph_dict.items()) if graph_dict[node].intersection(v)]
        )

        if subgraph:

            node_lst = [x for x in node_lst if x not in list(subgraph)]
            node_lst = node_lst
            result_dict[node] = subgraph

            result_dict = connect_graph(graph_dict, node_lst, result_dict)

        else:
            continue

        if len(node_lst) == 1 or len(node_lst) == 0:
            return result_dict
        break

    return result_dict


def run_grid_graph(df, result_lst, matrix, min_samples, min_silh, min_simi, max_dist):

    compare_lst = list()

    for label_lst in result_lst:

        pass_lst = build_pass_lst(label_lst, matrix, max_dist, min_silh)

        compare_lst += pass_lst

    total_comparisons = len(compare_lst)

    index_lst = lst_nums(0, total_comparisons - 1)

    connect_matrix = np.zeros((total_comparisons, total_comparisons))

    index_pairs = itertools.combinations(index_lst, 2)

    for i_index, j_index in index_pairs:

        result = check_connect(compare_lst, i_index, j_index, min_simi)

        i_index = result[0]
        j_index = result[1]
        status = result[2]

        connect_matrix[i_index, j_index] = status
        connect_matrix[j_index, i_index] = status

    node_lst = list(range(len(connect_matrix)))

    graph_dict = dict()
    result_dict = dict()

    for node in node_lst:

        neighbors = list(np.where(connect_matrix[:][node] == 1)[0])
        edges = neighbors + [node]
        graph_dict[node] = set(edges)

    result_dict = connect_graph(graph_dict, node_lst, result_dict)

    cluster_dict = dict()
    cluster_label = 0

    for node_lst in list(result_dict.values()):

        final_index_lst = set()

        for node in node_lst:

            index_lst = set(compare_lst[node][1])
            final_index_lst = final_index_lst.union(index_lst)

        if len(final_index_lst) >= min_samples:
            cluster_dict[cluster_label] = final_index_lst
            cluster_label = cluster_label + 1

    for cluster_label in list(cluster_dict.keys()):
        for index in cluster_dict[cluster_label]:
            df.at[index, cluster_col] = cluster_label

    if cluster_col in list(df.columns):
        df[cluster_col] = df[cluster_col].fillna(-1)
    else:
        df[cluster_col] = -1

    return df


def cluster_matrix(
    df,
    matrix,
    cluster_table_path,
    sum_table_path=None,
    report_table_path=None,
    eps_range="0.1-1.6",
    eps_step=0.1,
    min_samples_range="5-15",
    min_samples_step=1,
    silh_range=0.6,
    silh_step=0.1,
    simi_range=0.9,
    simi_step=0.1,
    max_dist=3.75,
    max_nn_dist=None,
    min_pdb=3,
    min_cf=1,
    min_min_samples=None,
    constr_matrix=None,
    max_constr_dist=None,
    merge_constr_dist=None,
    constr_method="mean",
    cluster_lim=None,
):

    eps_lst = build_range_lst(eps_range, eps_step, type=float)
    min_samples_lst = build_range_lst(min_samples_range, min_samples_step, type=int)
    silh_lst = build_range_lst(silh_range, silh_step, type=float)
    simi_lst = build_range_lst(simi_range, simi_step, type=float)

    if min_min_samples is None:
        min_min_samples = min(min_samples_lst)

    result_lst = list()
    for eps in tqdm(eps_lst, desc="Running Clustering", position=0, leave=True):
        for min_samples in min_samples_lst:
            result_lst.append(run_dbscan(matrix, eps, min_samples))

    run_lst = list()
    for min_silh in silh_lst:
        for min_simi in simi_lst:
            run_lst.append((min_silh, min_simi))

    if report_table_path is not None:
        report_df = pd.DataFrame()
        i = 0

    grid_graph_lst = list()
    cluster_count_lst = list()
    total_noise_lst = list()

    for run in tqdm(run_lst, desc="Running GridGraph", position=0, leave=True):

        min_silh = run[0]
        min_simi = run[1]

        df = run_grid_graph(
            df,
            result_lst,
            matrix,
            min_min_samples,
            min_silh,
            min_simi,
            max_dist,
        )

        if constr_matrix is not None:
            if merge_constr_dist is not None:
                if max_constr_dist is None:
                    max_constr_dist = build_cutoff_dict(
                        df, constr_matrix, method=constr_method
                    )
                df = merge_clusters(
                    df, constr_matrix, merge_constr_dist, merge_method=constr_method
                )

        result = prune_cluster_members(
            df,
            min_min_samples,
            matrix,
            max_nn_dist=max_nn_dist,
            constr_matrix=constr_matrix,
            max_constr_dist=max_constr_dist,
            constr_method=constr_method,
            min_pdb=min_pdb,
            min_cf=min_cf,
        )

        df = result[0]
        nn_pruned = result[1]
        total_pruned = nn_pruned
        if constr_matrix is not None:
            constr_pruned = result[2]
            total_pruned += constr_pruned

        cluster_df = mask_unequal(df, cluster_col, "Noise")

        total_cluster = len(cluster_df)

        cluster_count = len(lst_col(cluster_df, cluster_col, unique=True))

        count_dict = build_col_count_dict(df, cluster_col)
        total_noise = 0
        if "Noise" in list(count_dict.keys()):
            total_noise = count_dict["Noise"]

        cluster_count_lst.append(cluster_count)
        total_noise_lst.append(total_noise)
        grid_graph_lst.append((min_silh, min_simi))

        if report_table_path is not None:
            report_df.at[i, silh_col] = min_silh
            report_df.at[i, simi_col] = min_simi
            report_df.at[i, total_complete_col] = len(df)
            report_df.at[i, cluster_count_col] = cluster_count
            report_df.at[i, total_cluster_pre_col] = total_cluster + total_pruned
            report_df.at[i, total_cluster_post_col] = total_cluster
            report_df.at[i, total_noise_pre_col] = total_noise - total_pruned
            report_df.at[i, total_noise_post_col] = total_noise
            report_df.at[i, total_pruned_nn_col] = nn_pruned
            if constr_matrix is not None:
                report_df.at[i, total_pruned_constr_col] = constr_pruned

            i += 1

    if cluster_lim is not None:
        max_cluster = max(cluster_count_lst)
        for cluster_count in sort_lst(lst_unique(cluster_count_lst)):
            if cluster_count <= cluster_lim:
                max_cluster = cluster_count
            else:
                break

        max_cluster_index_lst = get_lst_val_indices(cluster_count_lst, max_cluster)
        min_noise = min([total_noise_lst[x] for x in max_cluster_index_lst])
        min_noise_index_lst = get_lst_val_indices(total_noise_lst, min_noise)

        best_run = max([x for x in min_noise_index_lst if x in max_cluster_index_lst])
    else:
        min_noise = min(total_noise_lst)
        min_noise_index_lst = get_lst_val_indices(total_noise_lst, min_noise)
        min_cluster = min([cluster_count_lst[x] for x in min_noise_index_lst])
        min_cluster_index_lst = get_lst_val_indices(cluster_count_lst, min_cluster)

        best_run = max([x for x in min_cluster_index_lst if x in min_noise_index_lst])

    if report_table_path is not None:
        report_df.at[best_run, select_col] = True
        report_df[select_col] = report_df[select_col].fillna(False)

    best_min_silh = grid_graph_lst[best_run][0]
    best_min_simi = grid_graph_lst[best_run][1]

    print(f"Best Silhouette={best_min_silh}; Best Similarity={best_min_simi}")

    df = run_grid_graph(
        df,
        result_lst,
        matrix,
        min_min_samples,
        best_min_silh,
        best_min_simi,
        max_dist,
    )

    pre_count = len(lst_col(df, cluster_col, unique=True))

    if constr_matrix is not None:
        if merge_constr_dist is not None:
            if max_constr_dist is None:
                max_constr_dist = build_cutoff_dict(
                    df, constr_matrix, method=constr_method
                )
            df = merge_clusters(
                df, constr_matrix, merge_constr_dist, merge_method=constr_method
            )

    post_merge_count = len(lst_col(df, cluster_col, unique=True))

    result = prune_cluster_members(
        df,
        min_min_samples,
        matrix,
        max_nn_dist=max_nn_dist,
        constr_matrix=constr_matrix,
        max_constr_dist=max_constr_dist,
        constr_method=constr_method,
        min_pdb=min_pdb,
        min_cf=min_cf,
    )

    df = result[0]

    post_prune_count = len(lst_col(df, cluster_col, unique=True))

    print(
        f"Total Clusters={pre_count-1}; Post-Merging={post_merge_count-1}; Post-Pruning={post_prune_count-1}"
    )

    save_table(cluster_table_path, df)

    if sum_table_path is not None:
        sum_df = build_sum_table(df)
        save_table(sum_table_path, sum_df)

    if report_table_path is not None:
        save_table(report_table_path, report_df)

    print("Clustered matrix!")


from src.download.modules import *
from src.download.lookfilesinside import look_what_is_inside
# IMPORTANT!!! Check your network connection: Wi-Fi or Wired Ethernet

# from src.download.catalogdownloader import catalog_downloader
# from src.download.latestcatreader import latest_catalog_reader
# from src.download.lookfilesinside import look_what_is_inside

# default_input_path_to_mmCIF = current_directory + "/mmCIF"
# default_input_path_to_mmCIF_assembly = current_directory + "/mmCIF_assembly"
# default_input_path_to_PDB = current_directory + "/PDB"
# default_input_path_to_PDB_assembly = current_directory + "/PDB_assembly"
# default_input_path_to_SIFTS = current_directory + "/SIFTS"
# default_output_path_to_mmCIF = current_directory + "/output_mmCIF"
# default_output_path_to_mmCIF_assembly = current_directory + "/output_mmCIF_assembly"
# default_output_path_to_PDB = current_directory + "/output_PDB"
# default_output_path_to_PDB_assembly = current_directory + "/output_PDB_assembly"


def download_pdb_assemblies_list_with_lxml():
    for _ in range(5):
        session = requests.Session()
        # rcsb = "https://files.rcsb.org/pub/pdb/data/biounit/PDB/all/"
        wwpdb = "https://ftp.wwpdb.org/pub/pdb/data/biounit/PDB/all/"
        links = set()
        try:
            with session.get(wwpdb, stream=True, timeout=600) as r:
                dom = html.fromstring(r.content)
                for link in dom.xpath('//a/@href'):
                    if ".gz" in link:
                        links.add(wwpdb + link)
            return links

        except requests.exceptions.RequestException as err:
            print(err)
            print("Will try again in 5 seconds...")
            time.sleep(5)


def url_formation_for_pool(format_to_download="mmCIF", list_of_file_names=(),
                           default_input_path_to_mmCIF=current_directory + "/mmCIF",
                           default_input_path_to_PDB=current_directory + "/PDB",
                           default_input_path_to_SIFTS=current_directory + "/SIFTS",
                           default_input_path_to_mmCIF_assembly=current_directory + "/mmCIF_assembly",
                           default_input_path_to_PDB_assembly=current_directory + "/PDB_assembly"):
    urls_to_target_files = list()
    for file_name in list_of_file_names:
        if file_name == "0000":
            continue
        if len(file_name) >= 4:
            if format_to_download == "mmCIF" or format_to_download == "all":
                if not os.path.exists(default_input_path_to_mmCIF):
                    os.makedirs(default_input_path_to_mmCIF)
                if "ent" in file_name and file_name.startswith('pdb'):
                    target_name = file_name[3:7] + ".cif.gz"
                else:
                    target_name = file_name[0:4] + ".cif.gz"
                urls_to_target_files.append("https://files.rcsb.org/pub/pdb/data/structures/all/mmCIF/" + target_name)

            if format_to_download == "PDB" or format_to_download == "all":
                if not os.path.exists(default_input_path_to_PDB):
                    os.makedirs(default_input_path_to_PDB)
                if "ent" in file_name and file_name.startswith('pdb'):
                    target_name = "pdb" + file_name[3:7] + ".ent.gz"
                else:
                    target_name = "pdb" + file_name[0:4] + ".ent.gz"
                urls_to_target_files.append("https://files.rcsb.org/pub/pdb/data/structures/all/pdb/" + target_name)

            if format_to_download == "SIFTS" or format_to_download == "all":
                if not os.path.exists(default_input_path_to_SIFTS):
                    os.makedirs(default_input_path_to_SIFTS)
                if "ent" in file_name and file_name.startswith('pdb'):
                    target_name = file_name[3:7] + ".xml.gz"
                else:
                    target_name = file_name[0:4] + ".xml.gz"
                urls_to_target_files.append("http://ftp.ebi.ac.uk/pub/databases/msd/sifts/xml/" + target_name)

            if format_to_download == "mmCIF_assembly" or format_to_download == "all":
                if not os.path.exists(default_input_path_to_mmCIF_assembly):
                    os.makedirs(default_input_path_to_mmCIF_assembly)
                if "ent" in file_name and file_name.startswith('pdb'):
                    target_name = file_name[3:7] + ".cif.gz"
                else:
                    target_name = file_name[0:4] + ".cif.gz"
                urls_to_target_files.append("https://www.ebi.ac.uk/pdbe/static/entry/" + target_name[:4] + "-assembly.xml")

            if format_to_download == "PDB_assembly":
                if not os.path.exists(default_input_path_to_PDB_assembly):
                    os.makedirs(default_input_path_to_PDB_assembly)
                urls_to_target_files.append("https://ftp.wwpdb.org/pub/pdb/data/biounit/PDB/all/" + file_name)

        else:
            raise ValueError("Input file names list is not correct!!! It cannot be less than 4 characters")

    return urls_to_target_files


def download_with_pool(urls_to_target_files=(),
                       default_input_path_to_mmCIF=current_directory + "/mmCIF",
                       default_input_path_to_PDB=current_directory + "/PDB",
                       default_input_path_to_SIFTS=current_directory + "/SIFTS",
                       default_input_path_to_mmCIF_assembly=current_directory + "/mmCIF_assembly",
                       default_input_path_to_PDB_assembly=current_directory + "/PDB_assembly"):
    for _ in range(3):
        try:
            file_name_start_pos = urls_to_target_files.rfind("/") + 1
            format_start_pos = file_name_start_pos - 4
            file_name = urls_to_target_files[file_name_start_pos:]
            format_of_db = urls_to_target_files[format_start_pos:format_start_pos + 3]

            r = requests.get(urls_to_target_files, stream=True, timeout=10)

            if format_of_db == "CIF":
                if r.status_code == requests.codes.ok:
                    with open(default_input_path_to_mmCIF + "/" + file_name, 'wb') as f:
                        for data in r:
                            f.write(data)

            if format_of_db == "pdb":
                if r.status_code == requests.codes.ok:
                    with open(default_input_path_to_PDB + "/" + file_name, 'wb') as f:
                        for data in r:
                            f.write(data)

            if format_of_db == "xml":
                if r.status_code == requests.codes.ok:
                    with open(default_input_path_to_SIFTS + "/" + file_name, 'wb') as f:
                        for data in r:
                            f.write(data)

            if format_of_db == "all":
                if r.status_code == requests.codes.ok:
                    with open(default_input_path_to_PDB_assembly + "/" + file_name, 'wb') as f:
                        for data in r:
                            f.write(data)

            if format_of_db == "try":
                if r.status_code == requests.codes.ok:
                    root = ET.fromstring(r.text)
                    for n in root:
                        compos_ID_list = list(n.attrib.items())
                        if compos_ID_list[1][0] == "id":
                            req_child = requests.get(
                                "https://www.ebi.ac.uk/pdbe/static/entry/" + file_name[0:4] + "-assembly-" + compos_ID_list[1][1] + ".cif.gz",
                                stream=True)
                            if req_child.status_code == requests.codes.ok:
                                with open(
                                        default_input_path_to_mmCIF_assembly + "/" + file_name[0:4] + "-assembly-" + compos_ID_list[1][1] + ".cif.gz",
                                        'wb') as f:
                                    for data in req_child:
                                        f.write(data)

        except requests.exceptions.RequestException as err:
            time.sleep(5)
            continue

        break


def run_downloads_with_ThreadPool(format_to_download="mmCIF", urls_to_target=(),
                                  default_input_path_to_mmCIF=current_directory + "/mmCIF",
                                  default_input_path_to_PDB=current_directory + "/PDB",
                                  default_input_path_to_SIFTS=current_directory + "/SIFTS",
                                  default_input_path_to_mmCIF_assembly=current_directory + "/mmCIF_assembly",
                                  default_input_path_to_PDB_assembly=current_directory + "/PDB_assembly"):
    for i in range(3):
        executor = ThreadPoolExecutor()
        partial_download_with_pool = partial(download_with_pool,
                                             default_input_path_to_mmCIF=default_input_path_to_mmCIF,
                                             default_input_path_to_PDB=default_input_path_to_PDB,
                                             default_input_path_to_SIFTS=default_input_path_to_SIFTS,
                                             default_input_path_to_mmCIF_assembly=default_input_path_to_mmCIF_assembly,
                                             default_input_path_to_PDB_assembly=default_input_path_to_PDB_assembly)

        jobs = [executor.submit(partial_download_with_pool, url) for url in urls_to_target]

        for _ in tqdm.tqdm(as_completed(jobs), total=len(jobs), miniters=1, position=0,
                           leave=True, desc="Downloading " + format_to_download + " files"):
            pass

        files_targeted = list()
        format_of_db = 0
        for url in urls_to_target:
            file_name_start_pos = url.rfind("/") + 1
            file_name = url[file_name_start_pos:]
            files_targeted.append(file_name)
            format_start_pos = file_name_start_pos - 4
            format_of_db = url[format_start_pos:format_start_pos + 3]

        if format_of_db == "CIF":
            input_files = look_what_is_inside("mmCIF", default_input_path_to_mmCIF=default_input_path_to_mmCIF)
        elif format_of_db == "pdb":
            input_files = look_what_is_inside('PDB', default_input_path_to_PDB=default_input_path_to_PDB)
        elif format_of_db == "xml":
            input_files = look_what_is_inside('SIFTS', default_input_path_to_SIFTS=default_input_path_to_SIFTS)
        elif format_of_db == "all":
            input_files = look_what_is_inside('PDB_assembly', default_input_path_to_PDB_assembly=default_input_path_to_PDB_assembly)
        elif format_of_db == "try":
            input_files = look_what_is_inside('mmCIF_assembly', default_input_path_to_mmCIF_assembly=default_input_path_to_mmCIF_assembly)
        else:
            input_files = set()

        # check_if_all_files_in = False
        #
        # for files_in in files_targeted:
        #     if files_in in input_files:
        #         pass
        #     else:
        #         check_if_all_files_in = True
        #
        # if check_if_all_files_in:
        #     urls_to_target = list(set(files_targeted) - set(input_files))c
        # else:
        #     break

        output_4char = set()
        for n in input_files:
            output_4char.add(n[:4])

        new_round_files_targeted = set()
        for n in files_targeted:
            if n[:4] in output_4char:
                continue
            else:
                new_round_files_targeted.add(n)
        files_targeted = new_round_files_targeted

        if len(files_targeted) == 0:
            break


# if __name__ == '__main__':
#     start = time.time()
#
#     download_all_PDB_assemblies = download_pdb_assemblies_list_with_lxml()
#     input_PDB_assembly_files_were_found = look_what_is_inside(
#         "PDB_assembly", default_input_path_to_PDB_assembly=default_input_path_to_PDB_assembly)
#     try:
#         len(download_all_PDB_assemblies)
#     except TypeError:
#         print("Cannot reach https://ftp.wwpdb.org/pub/pdb/data/biounit/PDB/all/ maybe try again later")
#
#     lefttodownload_PDB_assemblies = [n for n in download_all_PDB_assemblies if n.rsplit('/', 1)[-1] not in input_PDB_assembly_files_were_found]
#     print(lefttodownload_PDB_assemblies)
#     end = time.time()
#
#     print("Time consumed in working with wwpdb: ", end - start)

#     all_files = latest_catalog_reader()
#     all_mmCIF_files = all_files[0]
#     all_PDB_files = all_files[1]
#     all_SIFTS_files = all_files[2]
#
#     urls_to_target_mmCIF_files = url_formation_for_pool("mmCIF", all_mmCIF_files)
#     urls_to_target_PDB_files = url_formation_for_pool("PDB", all_PDB_files)
#     urls_to_target_SIFTS_files = url_formation_for_pool("SIFTS", all_SIFTS_files)
#
#     run_downloads_with_ThreadPool(urls_to_target_mmCIF_files)
#     run_downloads_with_ThreadPool(urls_to_target_PDB_files)
#     run_downloads_with_ThreadPool(urls_to_target_SIFTS_files)
#
#     download_pdb_assemblies = download_pdb_assemblies_list_with_lxml()
#     run_downloads_with_ThreadPool("pdb_assembly", download_pdb_assemblies)
#
#     urls_to_target_mmCIF_assembly_files = url_formation_for_pool("mmCIF_assembly", all_mmCIF_files)
#     run_downloads_with_ThreadPool("mmCIF_assembly", urls_to_target_mmCIF_assembly_files)



# -*- coding: utf-8 -*-
"""
  Copyright 2022 Mitchell Isaac Parker

   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.

"""

from ..functions.color import get_rgb, get_lst_colors
from ..functions.col import pdb_id_col, pocket_lig_col
from ..functions.lst import type_lst, lst_to_str, str_to_lst
from ..functions.table import lst_col
from ..functions.col import (
    core_path_col,
    pdb_code_col,
    modelid_col,
    chainid_col,
    bound_prot_chainid_col,
    bound_interf_chainid_col,
    interf_col,
    pocket_col,
    bio_lig_col,
    ion_lig_col,
    pharm_lig_col,
    chem_lig_col,
    mod_lig_col,
    mem_lig_col,
)
from ..functions.path import (
    append_file_path,
    modify_coord_path,
    get_file_name,
    get_eds_map_path,
    get_eds_diff_path,
    map_str,
    diff_str,
)
from ..functions.download import download_file
from ..functions.url import eds_url

polymer_color = "gray80"

bio_color = "white"
ion_color = "green"
pharm_color = "cyan"
chem_color = "magenta"
mod_color = "yellow"
mem_color = "salmon"
pocket_color = "orange"
prot_color = "purple"

hb_color = "tv_yellow"
wmhb_color = "tv_blue"

symbol_lst = [" ", "/", "_"]

pymol_color_dict = {
    bio_lig_col: bio_color,
    ion_lig_col: ion_color,
    pharm_lig_col: pharm_color,
    chem_lig_col: chem_color,
    mod_lig_col: mod_color,
    mem_lig_col: mem_color,
    pocket_lig_col: pocket_color,
    bound_prot_chainid_col: prot_color,
}


def init_pymol(pymol_file, color_dict=None):

    pymol_file.write("bg_color white\n")
    pymol_file.write("set antialias, 2\n")
    pymol_file.write("set orthoscopic, off\n")
    pymol_file.write("set depth_cue, 0\n")
    pymol_file.write("set sphere_scale, 0.4\n")
    pymol_file.write("set stick_radius, 0.2\n")
    pymol_file.write("set hash_max, 300\n")

    if color_dict is not None:
        for group, color in color_dict.items():
            pymol_file.write(f"set_color {group}_color, {color}\n")


def load_obj(pymol_file, path, obj, chainids=None):

    chainid_lst = type_lst(chainids)

    if ".pdb" in path or ".cif" in path:
        load_cmd = f"load {path}, {obj}"
    elif len(path) == 4:
        load_cmd = f"fetch {path}, {obj}"

    if chainid_lst is not None:
        chainids = lst_to_str(chainid_lst, join_txt="+")
        load_cmd += f"; remove model {obj} and not {chainids}//"

    pymol_file.write(f"{load_cmd}\n")


def init_obj(pymol_file, obj=None, ribbon=True, color_chainbow=False):

    color = polymer_color

    if ribbon:
        show = "ribbon"
        hide = "cartoon"
    else:
        show = "cartoon"
        hide = "ribbon"

    pymol_file.write("hide wire, solvent\n")

    if obj is None:
        pymol_file.write(f"show {show}\n")
        pymol_file.write(f"hide {hide}\n")
        pymol_file.write("hide lines\n")
        pymol_file.write("hide sticks\n")
        pymol_file.write("hide spheres\n")
        pymol_file.write(f"color {color}\n")
    else:
        pymol_file.write(f"show {show}, model {obj}\n")
        pymol_file.write(f"hide {hide}, model {obj}\n")
        pymol_file.write(f"hide lines, model {obj}\n")
        pymol_file.write(f"hide sticks, model {obj}\n")
        pymol_file.write(f"hide spheres, model {obj}\n")
        pymol_file.write(f"color {color}, model {obj}\n")

    if color_chainbow:
        pymol_file.write("util.chainbow\n")


def set_color(pymol_file, show_color, color, color_str, return_color=True):

    if show_color == True:
        color_str = color
    elif show_color is not None and show_color != False:
        color_rgb = get_rgb(show_color)
        pymol_file.write(f"set_color {color_str}, {color_rgb}\n")

    if return_color:
        return color_str


def color_sticks(pymol_file, stick_sele, stick_color=None, color_atomic=True):

    if stick_color is not None:
        pymol_file.write(f"color {stick_color}_color, {stick_sele}\n")
    if color_atomic:
        pymol_file.write(f"color atomic, {stick_sele} and (not elem C)\n")


def show_lig_structure(
    pymol_file,
    lig_col,
    lig_dict,
    show_lig,
    lig_color,
    group=None,
    spheres=False,
):

    if show_lig is not None and show_lig != False:

        if group is None:
            lig_str = lst_to_str(lig_dict[lig_col], join_txt="+")
        else:
            lig_str = lst_to_str(lig_dict[group][lig_col], join_txt="+")

        if lig_str is not None:
            if group is None:
                lig_sele = lig_col
                pymol_file.write(f"select {lig_sele}, resn {lig_str}\n")
            else:
                lig_sele = f"{group}_{lig_col}"
                pymol_file.write(f"select {lig_sele}, *_{group} and resn {lig_str}\n")

            pymol_file.write(f"color {lig_color}, {lig_sele}\n")

            if spheres:
                pymol_file.write(f"show spheres, {lig_sele}\n")
            else:
                pymol_file.write(f"show sticks, {lig_sele}\n")
                color_sticks(pymol_file, lig_sele)


def get_sup_sele(obj, chainid, sup_resids=None):

    sup_sele_lst = [f"model {obj} and chain {chainid}"]

    if sup_resids is not None:
        sup_resids = str(sup_resids)
        sele_resid_lst = sup_resids.replace(":", "+")
        sup_sele_lst.append(f"resi {lst_to_str(sele_resid_lst,join_txt='+')}")

    sup_sele_lst.append("bb.")
    sup_sele = lst_to_str(sup_sele_lst, join_txt=" and ")

    return sup_sele


def get_hb_sele(
    x_hb_resname, y_hb_resname, obj=None, x_hb_atomid_str=None, y_hb_atomid_str=None
):
    sele_lst = list()

    if obj is not None:
        sele_lst.append(obj)

    sele_lst.append(x_hb_resname)

    if x_hb_atomid_str is not None:
        sele_lst.append(x_hb_atomid_str.replace("+", "_"))

    sele_lst.append(y_hb_resname)

    if y_hb_atomid_str is not None:
        sele_lst.append(y_hb_atomid_str.replace("+", "_"))

    hb_sele_lst = sele_lst.copy()
    wmhb_sele_lst = sele_lst.copy()
    angle_sele_lst = sele_lst.copy()

    hb_sele_lst.append("hb")
    wmhb_sele_lst.append("wmhb")
    angle_sele_lst.append("angle")

    hb_sele = lst_to_str(hb_sele_lst, join_txt="_")
    wmhb_sele = lst_to_str(wmhb_sele_lst, join_txt="_")
    angle_sele = lst_to_str(angle_sele_lst, join_txt="_")

    return hb_sele, wmhb_sele, angle_sele


def write_pymol_script(
    pymol_df,
    pymol_pml_path,
    group_col=None,
    stick_resids=None,
    loop_resids=None,
    sup_resids=None,
    sup_coord_path=None,
    sup_chainid=None,
    sup_all=True,
    sup_group=False,
    group_order=None,
    style_ribbon=True,
    thick_bb=True,
    color_atomic=True,
    show_resids=None,
    cartoon_transp=0,
    surface_transp=1,
    show_bio=None,
    show_ion=None,
    show_pharm=None,
    show_chem=None,
    show_mod=None,
    show_mem=None,
    show_pocket=None,
    show_prot=None,
    color_palette=None,
    color_group=True,
    color_chainbow=False,
    prot_chainid_col=None,
    set_view=None,
    eds_dir=None,
    eds_resids=None,
    map_cutoff=1.0,
    diff_cutoff=3.0,
    carve_value=1.6,
    mesh_width=0.5,
    x_hb_resids=None,
    x_hb_atomids=None,
    y_hb_resids=None,
    y_hb_atomids=None,
    max_hb_dist=3.2,
    max_wmhb_dist=3.0,
    show_hb=True,
    show_wmhb=True,
    show_angle=False,
    add_h=False,
    coord_path_col=None,
    fetch_path=None,
):

    show_lig_dict = {
        bio_lig_col: show_bio,
        ion_lig_col: show_ion,
        pharm_lig_col: show_pharm,
        chem_lig_col: show_chem,
        mod_lig_col: show_mod,
        mem_lig_col: show_mem,
        pocket_lig_col: show_pocket,
    }

    df = pymol_df.reset_index(drop=True)

    if coord_path_col is None:
        coord_path_col = core_path_col

    if prot_chainid_col is None:
        prot_chainid_col = bound_prot_chainid_col

    stick_resid_lst = type_lst(stick_resids)
    loop_resid_lst = type_lst(loop_resids)

    x_hb_resid_lst = type_lst(x_hb_resids)
    x_hb_atomid_lst = type_lst(x_hb_atomids)
    y_hb_resid_lst = type_lst(y_hb_resids)
    y_hb_atomid_lst = type_lst(y_hb_atomids)

    if max_wmhb_dist is None:
        max_wmhb_dist = max_hb_dist

    index_lst = list(df.index.values)
    df_col_lst = list(df.columns)

    add_modelid = False
    if len(lst_col(df, modelid_col, unique=True)) > 1:
        add_modelid = True

    label_pdb = False
    if pdb_id_col in df_col_lst:
        label_pdb = True

    if coord_path_col == pdb_code_col:
        add_h = False

    if group_col is None:
        group_lst = index_lst
    else:
        for symbol in symbol_lst:
            df[group_col] = df[group_col].str.replace(symbol, "_")
        if group_order is not None:
            for symbol in symbol_lst:
                group_order = [x.replace(symbol, "_") for x in group_order]
            group_lst = group_order
        else:
            group_lst = lst_col(df, group_col, unique=True, return_str=True)

        group_dict = dict()
        for group in group_lst:
            group_dict[group] = list()

    if color_chainbow:
        color_group = False

    if (group_col is None or color_group is False) and loop_resids is not None:
        label_lst = loop_resid_lst.copy()
    else:
        label_lst = group_lst.copy()

    if type(color_palette) == dict:
        color_dict = dict()
        for key, val in color_palette.items():
            for symbol in symbol_lst:
                key = key.replace(symbol, "_")
            color_dict[key] = get_rgb(val)
    else:
        color_dict = get_lst_colors(
            label_lst, palette=color_palette, return_rgb=True, return_dict=True
        )

    append_file_path(pymol_pml_path)

    pymol_file = open(pymol_pml_path, "w")

    if fetch_path is not None:
        pymol_file.write(f"set fetch_path, {fetch_path}\n")

    init_pymol(pymol_file, color_dict=color_dict)

    show_color_dict = dict()
    for col, show in show_lig_dict.items():
        show_color_dict[col] = set_color(
            pymol_file, show, pymol_color_dict[col], f"{col}_color"
        )

    prot_color_str = set_color(pymol_file, show_prot, prot_color, "prot_color")

    if type(show_hb) == list:
        hb_color_lst = list()
        for i, color in enumerate(show_hb):
            hb_color_lst.append(set_color(pymol_file, color, hb_color, f"{i}_hb"))

    if type(show_wmhb) == list:
        wmhb_color_lst = list()
        for i, color in enumerate(show_wmhb):
            wmhb_color_lst.append(set_color(pymol_file, color, wmhb_color, f"{i}_wmhb"))

    show_lig_lst = list(show_lig_dict.keys())
    lig_dict = dict()

    if group_col is None:
        for col in show_lig_lst:
            lig_dict[col] = list()
    else:
        for group in group_lst:
            lig_dict[group] = dict()
            for col in show_lig_lst:
                lig_dict[group][col] = list()

    for index in index_lst:

        coord_path = df.at[index, coord_path_col]
        modelid = df.at[index, modelid_col]
        chainid = df.at[index, chainid_col]

        if add_h:
            coord_path = modify_coord_path(coord_path, return_pdb=True, add_h=True)

        obj_lst = list()

        if label_pdb:
            if coord_path_col == pdb_code_col:
                obj_lst.append(df.at[index, pdb_code_col])
                obj_lst.append(chainid)
            else:
                obj_lst.append(df.at[index, pdb_id_col])
            if add_modelid:
                obj_lst.append(modelid)
            if interf_col in df_col_lst:
                obj_lst.append(df.at[index, interf_col])
            if pocket_col in df_col_lst:
                obj_lst.append(df.at[index, pocket_col])
        else:
            obj_lst.append(get_file_name(coord_path))

        group = index
        if group_col is not None:
            group = df.at[index, group_col]
            obj_lst.append(group)

        obj = lst_to_str(obj_lst, join_txt="_")

        if group_col is not None:
            group_dict[str(group)].append([obj, coord_path, chainid])

        for col in show_lig_lst:
            if col in df_col_lst:
                ligs = df.at[index, col]
                if ligs != "None":
                    for lig in str_to_lst(ligs):
                        if lig != "GLY":
                            if group_col is None:
                                if lig not in lig_dict[col]:
                                    lig_dict[col].append(lig)
                            else:
                                if lig not in lig_dict[group][col]:
                                    lig_dict[group][col].append(lig)

        chainid_lst = type_lst(chainid)
        if show_prot is not None and show_prot != False:
            prot_chainid = df.at[index, prot_chainid_col]
            if prot_chainid != "None":
                prot_chainid_lst = str_to_lst(prot_chainid)
                chainid_lst += prot_chainid_lst

        load_obj(pymol_file, coord_path, obj, chainids=chainid_lst)

        chain_sele = f"model {obj} and chain {chainid}"

        if surface_transp < 1:
            pymol_file.write(f"show surface, {chain_sele}\n")

        if show_prot is not None and show_prot != False:
            if prot_chainid_col != bound_interf_chainid_col:
                if prot_chainid != "None":
                    prot_chainid_str = lst_to_str(prot_chainid_lst, join_txt="+")

                    prot_sele = f"{obj}_{prot_chainid_str}_{prot_chainid_col}"

                    pymol_file.write(
                        f"select {prot_sele}, model {obj} and chain {prot_chainid_str}\n"
                    )

    init_obj(pymol_file, ribbon=style_ribbon, color_chainbow=color_chainbow)

    if show_resids:
        hide_style = "cartoon"
        if style_ribbon:
            hide_style = "ribbon"
        pymol_file.write(f"hide {hide_style}, not resi {show_resids}\n")       

    if stick_resids is not None:
        for stick_resid in stick_resid_lst:
            stick_sele = f"{stick_resid}_stick"
            if group_col is None or color_group is False:
                pymol_file.write(
                    f"select {stick_sele}, resi {stick_resid} and not name O+N+C\n"
                )
                color_sticks(
                    pymol_file,
                    stick_sele,
                    color_atomic=color_atomic,
                )
            else:
                for group in group_lst:
                    group_stick_sele = f"{group}_{stick_sele}"
                    pymol_file.write(
                        f"select {group_stick_sele}, *_{group} and resi {stick_resid} and not name O+N+C\n"
                    )

                    color_sticks(
                        pymol_file,
                        group_stick_sele,
                        stick_color=group,
                        color_atomic=color_atomic,
                    )

                pymol_file.write(f"group {stick_sele}, *_{stick_sele}\n")

            pymol_file.write(f"show sticks, {stick_sele}\n")

    if surface_transp < 1:
        pymol_file.write(f"set surface_color, {polymer_color}, (all)\n")

    if loop_resids is not None:
        for loop_range in loop_resid_lst:

            loop_range = str(loop_range)

            resid_str = loop_range.replace(":", "+")
            loop_range = loop_range.replace(":", "_")

            loop_sele = f"{loop_range}_loop"

            if group_col is None or color_group is False:
                pymol_file.write(f"select {loop_sele}, resi {resid_str}\n")
                color_sticks(
                    pymol_file,
                    loop_sele,
                    stick_color=loop_range,
                    color_atomic=color_atomic,
                )

                if style_ribbon:
                    pymol_file.write(f"hide cartoon,{loop_sele}\n")
                    if thick_bb:
                        pymol_file.write(f"show sticks, bb. and {loop_sele}\n")

                if surface_transp < 1:
                    pymol_file.write(f"set surface_color, {loop_range}_color, {loop_sele}\n")

            else:
                for group in group_lst:
                    group_loop_sele = f"{group}_{loop_sele}"
                    pymol_file.write(
                        f"select {group_loop_sele}, *_{group} and resi {resid_str}\n"
                    )
                    color_sticks(
                        pymol_file,
                        group_loop_sele,
                        stick_color=group,
                        color_atomic=color_atomic,
                    )

                    if style_ribbon:
                        pymol_file.write(f"hide cartoon,{group_loop_sele}\n")
                        if thick_bb:
                            pymol_file.write(
                                f"show sticks, bb. and {group_loop_sele}\n"
                            )

                    if surface_transp < 1:
                        pymol_file.write(f"set surface_color, {group}_color, {group_loop_sele}\n")

                pymol_file.write(f"group {loop_sele}, *_{loop_sele}\n")

    if group_col is None:
        for col, show in show_lig_dict.items():
            spheres = False
            if col == ion_lig_col:
                spheres = True
            show_lig_structure(
                pymol_file,
                col,
                lig_dict,
                show,
                show_color_dict[col],
                spheres=spheres,
            )
    else:
        for group in group_lst:
            for col, show in show_lig_dict.items():
                spheres = False
                if col == ion_lig_col:
                    spheres = True
                show_lig_structure(
                    pymol_file,
                    col,
                    lig_dict,
                    show,
                    show_color_dict[col],
                    spheres=spheres,
                    group=group,
                )

        for col, show in show_lig_dict.items():
            if show is not None and show != False:
                pymol_file.write(f"group {col}, *_{col}\n")

    if show_prot is not None and show_prot != False:
        if prot_chainid_col != bound_interf_chainid_col:
            prot_style = "cartoon"
            if style_ribbon:
                prot_style = "ribbon"
            pymol_file.write(f"group {prot_chainid_col}, *_{prot_chainid_col}\n")
            pymol_file.write(f"show {prot_style}, {prot_chainid_col}\n")
            pymol_file.write(f"hide sticks, {prot_chainid_col}\n")
            pymol_file.write(f"color {prot_color_str}, {prot_chainid_col}\n")

    if len(df) > 1:
        if sup_all:
            if sup_coord_path is None:
                sup_coord_path = df.at[0, coord_path_col]
                if sup_chainid is None:
                    sup_chainid = df.at[0, chainid_col]

            if sup_coord_path is not None and sup_chainid is not None:

                sup_obj = "sup_obj"

                load_obj(
                    pymol_file,
                    sup_coord_path,
                    sup_obj,
                    chainids=sup_chainid,
                )

                sup_sele = get_sup_sele(
                    sup_obj,
                    sup_chainid,
                    sup_resids=sup_resids,
                )

                pymol_file.write(f"alignto {sup_sele}\n")
                pymol_file.write(f"center {sup_obj}\n")
                pymol_file.write(f"delete {sup_obj}\n")

    if group_col is not None:
        for group in group_lst:
            if sup_group:
                obj_val_lst = group_dict[group]

                if len(obj_val_lst) > 1:

                    sup_val_lst = obj_val_lst[0]

                    sup_obj = sup_val_lst[0]
                    sup_chainid = sup_val_lst[2]

                    sup_sele = get_sup_sele(
                        sup_obj,
                        sup_chainid,
                        sup_resids=sup_resids,
                    )

                    for val_lst in obj_val_lst:
                        group_obj = val_lst[0]
                        if group_obj != sup_obj:
                            pymol_file.write(f"align {group_obj}, {sup_sele}\n")

            pymol_file.write(f"group {group}, *_{group}\n")

    if x_hb_resid_lst is not None and y_hb_resid_lst is not None:

        pymol_file.write(f"set h_bond_cutoff_edge, {max_hb_dist}\n")

        for index in index_lst:

            coord_path = df.at[index, coord_path_col]
            modelid = df.at[index, modelid_col]
            chainid = df.at[index, chainid_col]

            obj_lst = list()

            if label_pdb:
                if coord_path_col == pdb_code_col:
                    obj_lst.append(df.at[index, pdb_code_col])
                    obj_lst.append(chainid)
                else:
                    obj_lst.append(df.at[index, pdb_id_col])
                if add_modelid:
                    obj_lst.append(modelid)
                if interf_col in df_col_lst:
                    obj_lst.append(df.at[index, interf_col])
                if pocket_col in df_col_lst:
                    obj_lst.append(df.at[index, pocket_col])
            else:
                obj_lst.append(get_file_name(coord_path))

            group = index
            if group_col is not None:
                group = df.at[index, group_col]
                obj_lst.append(group)

            obj = lst_to_str(obj_lst, join_txt="_")

            chain_sele = f"model {obj} and chain {chainid}"

            for i, (x_hb_resname, y_hb_resname) in enumerate(
                zip(x_hb_resid_lst, y_hb_resid_lst)
            ):

                if x_hb_atomid_lst is not None:
                    x_hb_atomid_str = lst_to_str(
                        type_lst(x_hb_atomid_lst[i]), join_txt="+"
                    )
                if y_hb_atomid_lst is not None:
                    y_hb_atomid_str = lst_to_str(
                        type_lst(y_hb_atomid_lst[i]), join_txt="+"
                    )

                hb_sele, wmhb_sele, angle_sele = get_hb_sele(
                    x_hb_resname,
                    y_hb_resname,
                    x_hb_atomid_str=x_hb_atomid_str,
                    y_hb_atomid_str=y_hb_atomid_str,
                    obj=obj,
                )

                x_res_str = "resi"
                y_res_str = "resi"

                if type(x_hb_resname) == str:
                    x_hb_resid = df.at[index, x_hb_resname]
                    x_res_str = "resname"
                else:
                    x_hb_resid = x_hb_resname

                if type(y_hb_resname) == str:
                    y_hb_resid = df.at[index, y_hb_resname]
                    y_res_str = "resname"
                else:
                    y_hb_resid = y_hb_resname

                if x_hb_resid is not None and y_hb_resid is not None:

                    x_hb_resid_sele = f"{chain_sele} and {x_res_str} {x_hb_resid}"
                    y_hb_resid_sele = f"{chain_sele} and {y_res_str} {y_hb_resid}"

                    if x_hb_atomid_lst[i] is not None:
                        x_hb_resid_sele += f" and name {x_hb_atomid_str}"

                    if y_hb_atomid_lst[i] is not None:
                        y_hb_resid_sele += f" and name {y_hb_atomid_str}"

                    write_hb = show_hb
                    if type(show_hb) == list:
                        write_hb = show_hb[i]

                    if write_hb is not False:
                        pymol_file.write(
                            f"dist {hb_sele}, {x_hb_resid_sele}, {y_hb_resid_sele}, cutoff={max_hb_dist}\n"
                        )

                    write_wmhb = show_wmhb
                    if type(show_wmhb) == list:
                        write_wmhb = show_wmhb[i]

                    if write_wmhb is not False:
                        wat_sele_1 = f"byres {chain_sele} and resname HOH within {max_wmhb_dist} of {x_hb_resid_sele}"
                        wat_sele_2 = f"byres {chain_sele} and resname HOH within {max_wmhb_dist} of {y_hb_resid_sele}"

                        wat_sele = f"({wat_sele_1}) in ({wat_sele_2})"

                        pymol_file.write(
                            f"dist {wmhb_sele}, {wat_sele}, ({x_hb_resid_sele})+({y_hb_resid_sele}), cutoff={max_wmhb_dist}\n"
                        )

                        if show_angle:
                            pymol_file.write(
                                f"angle {angle_sele},{x_hb_resid_sele}, {wat_sele}, {y_hb_resid_sele}\n"
                            )

                        pymol_file.write(f"sele HOH, {wat_sele}, merge=1\n")

    if x_hb_resids is not None and y_hb_resids is not None:

        pymol_file.write(f"set sphere_scale, 0.3, HOH\n")

        for i, (x_hb_resname, y_hb_resname) in enumerate(
            zip(x_hb_resid_lst, y_hb_resid_lst)
        ):

            if x_hb_atomid_lst is not None:
                x_hb_atomid_str = lst_to_str(type_lst(x_hb_atomid_lst[i]), join_txt="+")
            if y_hb_atomid_lst is not None:
                y_hb_atomid_str = lst_to_str(type_lst(y_hb_atomid_lst[i]), join_txt="+")

            hb_sele, wmhb_sele, angle_sele = get_hb_sele(
                x_hb_resname,
                y_hb_resname,
                x_hb_atomid_str=x_hb_atomid_str,
                y_hb_atomid_str=y_hb_atomid_str,
            )

            write_hb_status = show_hb
            write_hb_color = hb_color
            if type(show_hb) == list:
                write_hb_status = show_hb[i]
                write_hb_color = hb_color_lst[i]

            write_wmhb_status = show_hb
            write_wmhb_color = wmhb_color
            if type(show_wmhb) == list:
                write_wmhb_status = show_wmhb[i]
                write_wmhb_color = wmhb_color_lst[i]

            if group_col is None or color_group is False:
                if write_hb_status is not False:
                    pymol_file.write(f"group {hb_sele}, *_{hb_sele}\n")
                    pymol_file.write(f"color {write_hb_color}, {hb_sele}\n")
                if write_wmhb_status is not False:
                    pymol_file.write(f"group {wmhb_sele}, *_{wmhb_sele}\n")
                    pymol_file.write(f"color {write_wmhb_color}, {wmhb_sele}\n")
                    if show_angle:
                        pymol_file.write(f"group {angle_sele}, *_{angle_sele}\n")
                        pymol_file.write(f"color {write_wmhb_color}, {angle_sele}\n")
            else:
                for group in group_lst:
                    if write_hb_status is not False:
                        group_hb_sele = f"{group}_{hb_sele}"
                        pymol_file.write(f"group {group_hb_sele}, *_{group_hb_sele}\n")
                        if type(write_hb_status) == str:
                            group_hb_color = write_hb_color
                        else:
                            group_hb_color = f"{group}_color"
                        pymol_file.write(f"color {group_hb_color}, {group_hb_sele}\n")
                    if write_wmhb_status is not False:
                        group_wmhb_sele = f"{group}_{wmhb_sele}"
                        pymol_file.write(
                            f"group {group_wmhb_sele}, *_{group_wmhb_sele}\n"
                        )
                        if type(write_wmhb_status) == str:
                            group_wmhb_color = write_wmhb_color
                        else:
                            group_wmhb_color = f"{group}_color"
                        pymol_file.write(
                            f"color {group_wmhb_color}, {group_wmhb_sele}\n"
                        )
                        if show_angle:
                            group_angle_sele = f"{group}_{angle_sele}"
                            pymol_file.write(
                                f"group {group_angle_sele}, *_{group_angle_sele}\n"
                            )
                            pymol_file.write(
                                f"color {group_wmhb_color}, {group_angle_sele}\n"
                            )

            pymol_file.write(f"set dash_width, 3\n")

            if show_angle:
                pymol_file.write(f"hide angles\n")

            pymol_file.write(f"show spheres, HOH\n")
            pymol_file.write(f"color red, HOH\n")

    if set_view is not None:
        pymol_file.write(f"set_view {set_view}\n")

    if eds_dir is not None and eds_resids is not None:
        if len(df) == 1:
            if pdb_code_col in df_col_lst:
                pdb_code = df.at[0, pdb_code_col]
                chainid = df.at[0, chainid_col]

                map_eds_url = f"{eds_url}{pdb_code}.ccp4"
                diff_eds_url = f"{eds_url}{pdb_code}_diff.ccp4"

                map_path = get_eds_map_path(pdb_code, dir_path=eds_dir)
                diff_path = get_eds_diff_path(pdb_code, dir_path=eds_dir)

                download_file(map_eds_url, map_path)
                download_file(diff_eds_url, diff_path)

                pymol_file.write(f"load {map_path}, {map_str}\n")
                pymol_file.write(f"load {diff_path}, {diff_str}\n")

                mesh_map_sele = f"mesh_{map_str}"
                mesh_diff_sele = f"mesh_{diff_str}"

                eds_resids = str(eds_resids)

                eds_str = eds_resids.replace(":", "+")

                eds_sele = f"{obj} and chain and resi {eds_str}"

                pymol_file.write(
                    f"isomesh {mesh_map_sele}, {map_str}, {map_cutoff}, {eds_sele}, carve={carve_value}\n"
                )
                pymol_file.write(
                    f"isomesh {mesh_diff_sele}, {diff_str}, {diff_cutoff}, {eds_sele}\n"
                )

                pymol_file.write(f"color gray50, {mesh_map_sele}\n")
                pymol_file.write(f"color green, {mesh_diff_sele}\n")
                pymol_file.write("set mesh_negative_color, red\n")
                pymol_file.write("set mesh_negative_visible\n")

                pymol_file.write(f"orient {eds_sele}\n")

                pymol_file.write(f"set mesh_width, {mesh_width}\n")

    if not add_h:
        pymol_file.write("remove hydrogens\n")

    pymol_file.write(f"set cartoon_transparency, {cartoon_transp}\n")

    pymol_file.write(f"set transparency, {surface_transp}\n")

    pymol_file.write("hide labels\n")

    pymol_file.write("zoom center, 25\n")

    pymol_file.close()

    print("Wrote PyMOL script!")
